#include "parallel_finder.h"

#include <mutex>
#include <thread>

using std::mutex, std::thread, std::lock_guard, std::vector, std::forward_list,
    std::string, std::string_view;

ParallelFinder::ParallelFinder(const WordList &word_list)
    : word_list_(word_list) {}

forward_list<const string *>
ParallelFinder::find_prefix(string_view search_term) const {
  forward_list<const string *> result;
  mutex result_mutex;

  const auto word_list_size = word_list_.size();

  const size_t thread_count = thread::hardware_concurrency();

  vector<thread> threads;
  for (size_t thread_index = 0; thread_index < thread_count; ++thread_index) {
    const size_t first_word_index =
        thread_index * (word_list_size / thread_count);
    const size_t last_word_index =
        (thread_index == thread_count - 1)
            ? word_list_size
            : (thread_index + 1) * (word_list_size / thread_count);

    threads.emplace_back(
        [](const WordList &word_list, const string_view &search_term,
           forward_list<const string *> &result, size_t start_index,
           size_t end_index, mutex &result_mutex) {
          forward_list<const string *> thread_results;
          for (size_t index = start_index; index < end_index; ++index) {
            const auto &current_word = word_list[index];
            if (current_word.starts_with(search_term)) {
              thread_results.push_front(&current_word);
            }
          }
          if (!thread_results.empty()) {
            const lock_guard<mutex> lock(result_mutex);
            result.merge(thread_results);
          }
        },
        cref(word_list_), cref(search_term), ref(result), first_word_index,
        last_word_index, ref(result_mutex));
  }

  for (auto &thread : threads) {
    thread.join();
  }

  return result;
}
